from unsilence import Unsilence

u = Unsilence("[input_file]")

u.detect_silence()

estimated_time = u.estimate_time(audible_speed=5, silent_speed=2)  # Estimate time savings
print(estimated_time)

u.render_media("[output_file]")  # No options specified
u.render_media("[output_file]", audible_speed=2, silent_speed=8)  # Speed options
u.render_media("[output_file]", audible_volume=2, silent_volume=0)  # Volume options
u.render_media("[output_file]", audio_only=True)  # Audio only specified



import sys

from rich.console import Console
from rich.progress import Progress

from unsilence.Unsilence import Unsilence
from unsilence.command_line.ChoiceDialog import choice_dialog
from unsilence.command_line.ParseArguments import parse_arguments
from unsilence.command_line.PrettyTimeEstimate import pretty_time_estimate
from unsilence.command_line.TerminalSupport import repair_console


def main():
    """
    Entry Point if this script is run as a script instead of a library
    :return: None
    """
    try:
        repair_console()
        run()
    except KeyboardInterrupt:
        print("\nInterrupted")

    repair_console()
    sys.exit(0)


def run():
    """
    Run the Console Interface for Unsilence
    :return: None
    """
    sys.tracebacklimit = 0

    args = parse_arguments()
    console = Console()

    if args.debug:
        sys.tracebacklimit = 1000

    if args.output_file.exists() and not args.non_interactive_mode:
        if not choice_dialog(console, "File already exists. Overwrite?", default=False):
            return

    args_dict = vars(args)

    argument_list_for_silence_detect = [
        "silence_level", "silence_time_threshold", "short_interval_threshold", "stretch_time"
    ]

    argument_dict_for_silence_detect = {
        key: args_dict[key] for key in argument_list_for_silence_detect if key in args_dict.keys()
    }

    argument_list_for_renderer = [
        "audio_only", "audible_speed", "silent_speed", "audible_volume", "silent_volume",
        "drop_corrupted_intervals", "threads"
    ]

    argument_dict_for_renderer = {
        key: args_dict[key] for key in argument_list_for_renderer if key in args_dict.keys()
    }

    progress = Progress()

    continual = Unsilence(args.input_file)

    with progress:
        def update_task(current_task):
            def handler(current_val, total):
                progress.update(current_task, total=total, completed=current_val)

            return handler

        silence_detect_task = progress.add_task("Calculating Intervals...", total=1)

        continual.detect_silence(
            on_silence_detect_progress_update=update_task(silence_detect_task),
            **argument_dict_for_silence_detect
        )

        progress.stop()
        progress.remove_task(silence_detect_task)

        print()

        estimated_time = continual.estimate_time(args.audible_speed, args.silent_speed)
        console.print(pretty_time_estimate(estimated_time))

        print()

        if not args.non_interactive_mode:
            if not choice_dialog(console, "Continue with these options?", default=True):
                return

        progress.start()
        rendering_task = progress.add_task("Rendering Intervals...", total=1)
        concat_task = progress.add_task("Combining Intervals...", total=1)

        continual.render_media(
            args.output_file,
            on_render_progress_update=update_task(rendering_task),
            on_concat_progress_update=update_task(concat_task),
            **argument_dict_for_renderer
        )

        progress.stop()

    console.print("\n[green]Finished![/green] :tada:")
    print()



import datetime

from rich.table import Table


def format_timedelta(seconds):
    """
    Generates a pretty time representation for seconds (format: hour:minute:second)
    :param seconds: Amount of seconds (can be negative)
    :return: String representation
    """
    if seconds < 0:
        return f"-{datetime.timedelta(seconds=-seconds)}"
    else:
        return str(datetime.timedelta(seconds=seconds))


def pretty_time_estimate(time_data: dict):
    """
    Generates a rich.table.Table object from the time_data dict (from lib.Intervals.TimeCalculations.calculate_time)
    :param time_data: time_data dict (from lib.Intervals.TimeCalculations.calculate_time)
    :return: rich.table.Table object
    """
    table = Table(show_header=True, header_style="bold magenta")
    table.add_column("Type")
    table.add_column("Before")
    table.add_column("After")
    table.add_column("Difference")

    reorderer_time_data = {"all": {}, "audible": {}, "silent": {}}
    for column, row_with_values in time_data.items():
        for row, values in row_with_values.items():
            time_delta = format_timedelta(round(values[0]))
            reorderer_time_data[row][column] = f"{time_delta} ([cyan]{round(values[1] * 100, 1)}%[/cyan])"

    table.add_row(
        "Combined",
        reorderer_time_data["all"]["before"],
        reorderer_time_data["all"]["after"],
        reorderer_time_data["all"]["delta"]
    )

    table.add_row(
        "Audible",
        reorderer_time_data["audible"]["before"],
        reorderer_time_data["audible"]["after"],
        reorderer_time_data["audible"]["delta"]
    )

    table.add_row(
        "Silent",
        reorderer_time_data["silent"]["before"],
        reorderer_time_data["silent"]["after"],
        reorderer_time_data["silent"]["delta"]
    )

    return table



import atexit
import shutil
from pathlib import Path

from unsilence.lib.detect_silence.DetectSilence import detect_silence
from unsilence.lib.intervals.Intervals import Intervals
from unsilence.lib.intervals.TimeCalculations import calculate_time
from unsilence.lib.render_media.MediaRenderer import MediaRenderer
from unsilence.lib.tools.ffmpeg_version import get_ffmpeg_version


class Unsilence:
    """
    Unsilence Class to remove (or isolate or many other use cases) silence from audible video parts
    """

    def __init__(self, input_file: Path, temp_dir: Path = Path(".tmp")):
        """
        Initializes a new Unsilence Obqject
        :param input_file: The file that should be processed
        :param temp_dir: The temp dir where temporary files can be saved
        """
        self.__input_file = Path(input_file)
        self.__temp_dir = Path(temp_dir)
        self.__intervals: Intervals = None

        ffmpeg_version = get_ffmpeg_version()
        if ffmpeg_version is None:
            raise EnvironmentError("ffmpeg not found!")
        else:
            if ffmpeg_version[0] < 4 or ffmpeg_version[1] < 2:
                raise EnvironmentError(
                    "ffmpeg with version {} found, but a version > 4.2.0 is required!".format(
                        "".join([str(x) for x in ffmpeg_version])
                    )
                )

        atexit.register(self.cleanup)

    def detect_silence(self, **kwargs):
        """
        Detects silence of the file (Options can be specified in kwargs)
        :param kwargs: Keyword Args, more information below
        :return: A generated Intervals object

        kwargs:
            silence_level: Threshold of what should be classified as silent/audible (default -35) (in dB)
            silence_time_threshold: Resolution of the ffmpeg detection algorithm (default 0.5) (in seconds)
            short_interval_threshold : The shortest allowed interval length (default: 0.3) (in seconds)
            stretch_time: Time the interval should be enlarged/shrunken (default 0.25) (in seconds)
            on_silence_detect_progress_update: Function that should be called on progress update
                (called like: func(current, total))
        """
        self.__intervals = detect_silence(self.__input_file, **kwargs)
        return self.__intervals

    def set_intervals(self, intervals: Intervals):
        """
        Set the intervals so that they do not need to be re-detected
        :param intervals: Intervals collection
        :return: None
        """
        self.__intervals = intervals

    def get_intervals(self):
        """
        Get the current Intervals so they can be reused if wanted
        :return: Intervals collection
        """
        return self.__intervals

    def estimate_time(self, audible_speed: float = 6, silent_speed: float = 1):
        """
        Estimates the time (savings) when the current options are applied to the intervals
        :param audible_speed: The speed at which the audible intervals get played back at
        :param silent_speed: The speed at which the silent intervals get played back at
        :return: Dictionary of time information
        :exception: ValueError if silence detection was never run
        """
        if self.__intervals is None:
            raise ValueError("Silence detection was not yet run and no intervals where given manually!")

        return calculate_time(self.__intervals, audible_speed, silent_speed)

    def render_media(self, output_file: Path, **kwargs):
        """
        Renders the current intervals with options specified in the kwargs
        :param output_file: Where the final file should be saved at
        :param kwargs: Keyword args, more information below
        :return: None

        kwargs:
            audio_only: Whether the output should be audio only (bool)
            audible_speed: The speed at which the audible intervals get played back at (float)
            silent_speed: The speed at which the silent intervals get played back at (float)
            audible_volume: The volume at which the audible intervals get played back at (float)
            silent_volume: The volume at which the silent intervals get played back at (float)
            drop_corrupted_intervals: Whether corrupted video intervals should be discarded or tried to recover (bool)
            threads: Number of threads to render simultaneously (int > 0)
            on_render_progress_update: Function that should be called on render progress update
                (called like: func(current, total))
            on_concat_progress_update: Function that should be called on concat progress update
                (called like: func(current, total))
        """
        if self.__intervals is None:
            raise ValueError("Silence detection was not yet run and no intervals where given manually!")

        renderer = MediaRenderer(self.__temp_dir)
        renderer.render(self.__input_file, output_file, self.__intervals, **kwargs)

    def cleanup(self):
        """
        Cleans up the temporary directories, called automatically when the program ends
        :return:
        """
        if self.__temp_dir.exists():
            shutil.rmtree(self.__temp_dir)



import argparse
from pathlib import Path


def convert_to_path(should_exist=True, should_parents_exist=True):
    """
    Sets options for the nested class
    :param should_exist: If the file needs to exists
    :param should_parents_exist: If the files parents need to exist
    :return: Nested Handle Function
    """

    def handle(s):
        """
        Checks whether the string s fulfils the options set earlier
        :param s: Input string
        :return: Path Object or None
        """
        path = Path(s).absolute()
        if not path.exists() and should_exist:
            raise FileNotFoundError("File does not exist")

        if not path.parent.exists() and should_parents_exist:
            raise IOError("Parent directory not found")

        return path

    return handle


def number_bigger_than_zero(s):
    """
    Returns the Number representation of s if it is bigger than zero, else an error occurs
    :param s: Input string
    :return: Integer or None
    """
    i = int(s)

    if i <= 0:
        raise ValueError("Value must be larger than 0")

    return i


def parse_arguments():
    """
    Parses console arguments for the Unsilence Console Interface
    :return: List of Console Line Arguments
    """
    parser = argparse.ArgumentParser(
        description="Remove silence from media files",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter
    )

    parser.add_argument("input_file", type=convert_to_path(should_exist=True),
                        help="Path to the file that contains silence")
    parser.add_argument("output_file", type=convert_to_path(should_exist=False, should_parents_exist=True),
                        help="Path to where the finished media file should be")

    parser.add_argument("-ao", "--audio-only", action="store_true",
                        help="Whether the output should not contain a video channel")

    parser.add_argument("-as", "--audible-speed", type=float, default=1,
                        help="The speed at which audible parts should be played back at")
    parser.add_argument("-ss", "--silent-speed", type=float, default=6,
                        help="The speed at which silent parts should be played back at")
    parser.add_argument("-av", "--audible-volume", type=float, default=1,
                        help="The volume at which audible parts should be played back at")
    parser.add_argument("-sv", "--silent-volume", type=float, default=0.5,
                        help="The volume at which silent parts should be played back at")
    parser.add_argument("-dci", "--drop-corrupted-intervals", action="store_true",
                        help="Whether corrupted video intervals should be discarded or tried to recover")

    parser.add_argument("-t", "--threads", type=number_bigger_than_zero, default=2,
                        help="Number of threads to be used while rendering")
    parser.add_argument("-sl", "--silence-level", type=float, default=-35,
                        help="Minimum volume in decibel to be classified as audible")
    parser.add_argument("-stt", "--silence-time-threshold", type=float, default=0.5,
                        help="Resolution of the silence detection (seconds)")
    parser.add_argument("-sit", "--short-interval-threshold", type=float, default=0.3,
                        help="Intervals smaller than this value (seconds) get combined into a larger interval")
    parser.add_argument("-st", "--stretch-time", type=float, default=0.25,
                        help="Time (seconds) that should be added to audible intervals and removed from silent "
                             "intervals")

    parser.add_argument("-y", "--non-interactive-mode", action="store_true",
                        help="Always answers yes if a dialog would show up")

    parser.add_argument("-d", "--debug", action="store_true",
                        help="Enable debug output (StackTrace)")

    return parser.parse_args()



import re
import subprocess
from pathlib import Path

from unsilence.lib.intervals.Intervals import Intervals, Interval


def detect_silence(input_file: Path, **kwargs):
    """
    Detects silence in a file and outputs the intervals (silent/not silent) as a lib.Intervals.Intervals object
    :param input_file: File where silence should be detected
    :param kwargs: Various Parameters, see below
    :return: lib.Intervals.Intervals object

    kwargs:
        silence_level: Threshold of what should be classified as silent/audible (default -35) (in dB)
        silence_time_threshold: Resolution of the ffmpeg detection algorithm (default 0.5) (in seconds)
        short_interval_threshold : The shortest allowed interval length (default: 0.3) (in seconds)
        stretch_time: Time the interval should be enlarged/shrunken (default 0.25) (in seconds)
        on_silence_detect_progress_update: Function that should be called on progress update
            (called like: func(current, total))
    """
    input_file = Path(input_file).absolute()

    if not input_file.exists():
        raise FileNotFoundError(f"Input file {input_file} does not exist!")

    silent_detect_progress_update = kwargs.get("on_silence_detect_progress_update", None)

    command = [
        "ffmpeg",
        "-i", str(input_file),
        "-af",
        f"silencedetect=noise={kwargs.get('silence_level', -35)}dB:d={kwargs.get('silence_time_threshold', 0.5)}",
        "-f", "null",
        "-"
    ]

    console_output = subprocess.Popen(
        command,
        stdout=subprocess.PIPE,
        stderr=subprocess.STDOUT,
        universal_newlines=True
    ).stdout

    intervals = Intervals()
    current_interval = Interval(start=0, end=0, is_silent=False)
    media_duration = None

    for line in console_output:
        if "[silencedetect" in line:
            capture = re.search("\\[silencedetect @ [0-9xa-f]+] silence_([a-z]+): (-?[0-9]+.?[0-9]*[e-]*[0-9]*)",
                                line)

            event = capture[1]
            time = float(capture[2])

            if silent_detect_progress_update is not None:
                silent_detect_progress_update(time, media_duration)

            if event == "start":
                if current_interval.start != time:
                    current_interval.end = time
                    intervals.add_interval(current_interval)
                current_interval = Interval(start=time, is_silent=True)

            if event == "end":
                current_interval.end = time
                intervals.add_interval(current_interval)
                current_interval = Interval(start=time, is_silent=False)

        elif "Duration" in line:
            capture = re.search("Duration: ([0-9:]+.?[0-9]*)", line)
            hour, minute, second_millisecond = capture[1].split(":")
            second, millisecond = second_millisecond.split(".")
            media_duration = float(str(int(second) + 60 * (int(minute) + 60 * int(hour))) + "." + millisecond)

    current_interval.end = media_duration
    intervals.add_interval(current_interval)

    if silent_detect_progress_update is not None:
        silent_detect_progress_update(media_duration, media_duration)

    intervals.optimize(
        kwargs.get('short_interval_threshold', 0.3),
        kwargs.get('stretch_time', 0.25)
    )

    return intervals



import re
import subprocess


def get_ffmpeg_version():
    try:
        console_output = subprocess.run(
            ["ffmpeg", "-version"],
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT,
        ).stdout
    except FileNotFoundError:
        return None

    regex = r"ffmpeg version \D?(\d+(?:\.\d+)*)"
    match = re.search(regex, str(console_output))

    groups = match.groups()
    version = groups[0].split(".")

    return [int(version_part) for version_part in version]



import setuptools

with open("README.md", "r") as f:
    long_description = f.read()

with open('requirements.txt') as f:
    requirements = f.read().splitlines()

setuptools.setup(
    name='unsilence',
    version='1.0.6',
    install_requires=requirements,
    license='MIT License',
    author='Tim-Luca Lagmöller',
    author_email='mail@lagmoellertim.de',
    description='Remove Silence from Media Files',
    long_description=long_description,
    long_description_content_type="text/markdown",
    url='https://github.com/lagmoellertim/unsilence',
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    entry_points={
        "console_scripts": ["unsilence=unsilence.command_line.EntryPoint:main"]
    }
)



from unsilence.lib.intervals.Intervals import Intervals


def calculate_time(intervals: Intervals, audible_speed: float, silent_speed: float):
    """
    Generates a time estimate on the time saved if the current speed settings get applied
    :param intervals: Intervals which should be estimated (lib.Intervals.Intervals)
    :param audible_speed: The speed at which audible intervals should be played back at
    :param silent_speed: The speed at which silent intervals should be played back at
    :return: Time calculation dict
    """
    time_data = {"before": {}, "after": {}, "delta": {}}
    audible = 0
    silent = 0
    for interval in intervals.intervals:
        if interval.is_silent:
            silent += interval.duration
        else:
            audible += interval.duration

    time_data["before"]["all"] = (audible + silent, 1)
    time_data["before"]["audible"] = (audible, audible / time_data["before"]["all"][0])
    time_data["before"]["silent"] = (silent, silent / time_data["before"]["all"][0])

    new_audible = audible / audible_speed
    new_silent = silent / silent_speed

    time_data["after"]["all"] = (new_audible + new_silent, (new_audible + new_silent) / time_data["before"]["all"][0])
    time_data["after"]["audible"] = (new_audible, new_audible / time_data["before"]["all"][0])
    time_data["after"]["silent"] = (new_silent, new_silent / time_data["before"]["all"][0])

    time_data["delta"]["all"] = (
        time_data["after"]["all"][0] - time_data["before"]["all"][0],
        time_data["after"]["all"][1] - time_data["before"]["all"][1]
    )

    time_data["delta"]["audible"] = (
        time_data["after"]["audible"][0] - time_data["before"]["audible"][0],
        time_data["after"]["audible"][1] - time_data["before"]["audible"][1]
    )

    time_data["delta"]["silent"] = (
        time_data["after"]["silent"][0] - time_data["before"]["silent"][0],
        time_data["after"]["silent"][1] - time_data["before"]["silent"][1]
    )

    # Timedelta not working like i want

    return time_data



class Interval:
    """
    Represents a section in time where the media file is either silent or audible
    """

    def __init__(self, start=0, end=0, is_silent=False):
        """
        Initializes an Interval object
        :param start: Start time of the interval in seconds
        :param end: End time of the interval in seconds
        :param is_silent: Whether the interval is silent or not
        """
        self.__start = start
        self.__end = end
        self.__duration = self.__end - self.__start
        self.is_silent = is_silent

    @property
    def start(self):
        """
        Get the start time
        :return: start time in seconds
        """
        return self.__start

    @start.setter
    def start(self, new_start):
        """
        Sets the new start time and updates the duration
        :param new_start: start time in seconds
        :return: None
        """
        self.__start = new_start
        self.__duration = self.__end - self.__start

    @property
    def end(self):
        """
        Get the end time
        :return: end time in seconds
        """
        return self.__end

    @end.setter
    def end(self, new_end):
        """
        Sets the new end time and updates the duration
        :param new_end: start time in seconds
        :return: None
        """
        self.__end = new_end
        self.__duration = self.__end - self.__start

    @property
    def duration(self):
        """
        Returns the duration of the interval
        :return: Duration of the interval
        """
        return self.__duration

    def enlarge_audible_interval(self, stretch_time, is_start_interval=False, is_end_interval=False):
        """
        Enlarges/Shrinks the audio interval, based on if it is silent or not
        :param stretch_time: Time the interval should be enlarged/shrunken
        :param is_start_interval: Whether the current interval is at the start (should not enlarge/shrink)
        :param is_end_interval: Whether the current interval is at the end (should not enlarge/shrink)
        :return: None
        """
        if stretch_time >= self.duration:
            raise Exception("Stretch time to large, please choose smaller size")

        stretch_time_part = (-1 if self.is_silent else 1) * stretch_time / 2

        if not is_start_interval:
            self.start -= stretch_time_part

        if not is_end_interval:
            self.end += stretch_time_part

    def copy(self):
        """
        Creates a deep copy of this Interval
        :return: Interval deepcopy
        """
        return Interval(self.start, self.end, self.is_silent)

    def serialize(self):
        """
        Serializes the current interval into a dict format
        :return: serialized dict
        """
        return {"start": self.start, "end": self.end, "is_silent": self.is_silent}

    @staticmethod
    def deserialize(serialized_obj: dict):
        """
        Deserializes a previously serializes Interval and generates a new Interval with this data
        :param serialized_obj: previously serializes Interval (type dict)
        :return: Interval
        """
        return Interval(serialized_obj["start"], serialized_obj["end"], serialized_obj["is_silent"])

    def __repr__(self):
        """
        String representation
        :return: String representation
        """
        return f"<Interval start={self.start} end={self.end} duration={self.duration} is_silent={self.is_silent}>"



import pathlib
import queue
import subprocess
import threading
from types import SimpleNamespace

from unsilence.lib.intervals.Interval import Interval


class RenderIntervalThread(threading.Thread):
    """
    Worker thread that can render/process intervals based on defined options
    """

    def __init__(self, thread_id, input_file: pathlib.Path, render_options: SimpleNamespace, task_queue: queue.Queue,
                 thread_lock: threading.Lock, **kwargs):
        """
        Initializes a new Worker (is run in daemon mode)
        :param thread_id: ID of this thread
        :param input_file: The file the worker should work on
        :param render_options: The parameters on how the video should be processed, more details below
        :param task_queue: A queue object where the worker can get more tasks
        :param thread_lock: A thread lock object to acquire and release thread locks
        :param kwargs: Keyword Args, see below for more information
        """
        super().__init__()
        self.daemon = True
        self.thread_id = thread_id
        self.task_queue = task_queue
        self.thread_lock = thread_lock
        self.__should_exit = False
        self.__input_file = input_file
        self.__on_task_completed = kwargs.get("on_task_completed", None)
        self.__render_options = render_options

    def run(self):
        """
        Start the worker. Worker runs until stop() is called. It runs in a loop, takes a new task if available, and
        processes it
        :return: None
        """
        while not self.__should_exit:
            self.thread_lock.acquire()

            if not self.task_queue.empty():
                task: SimpleNamespace = self.task_queue.get()
                self.thread_lock.release()

                completed = self.__render_interval(
                    task.interval_output_file,
                    task.interval,
                    drop_corrupted_intervals=self.__render_options.drop_corrupted_intervals
                )

                if self.__on_task_completed is not None:
                    self.__on_task_completed(task, not completed)
            else:
                self.thread_lock.release()

    def stop(self):
        """
        Stops the worker after its current task is finished
        :return:
        """
        self.__should_exit = True

    def __render_interval(self, interval_output_file: pathlib.Path, interval: Interval,
                          apply_filter=True, drop_corrupted_intervals=False):
        """
        Renders an interval with the given render options
        :param interval_output_file: Where the current output file should be saved
        :param interval: The current Interval that should be processed
        :param apply_filter: Whether the AV-Filter should be applied or if the media interval should be left untouched
        :param drop_corrupted_intervals: Whether to remove corrupted frames from the video or keep them in unedited
        :return: Whether it is corrupted or not
        """

        command = self.__generate_command(interval_output_file, interval, apply_filter)

        console_output = subprocess.run(
            command,
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT,
        )

        if "Conversion failed!" in str(console_output.stdout).splitlines()[-1]:
            if drop_corrupted_intervals:
                return False
            if apply_filter:
                self.__render_interval(
                    interval_output_file,
                    interval,
                    apply_filter=False,
                    drop_corrupted_intervals=drop_corrupted_intervals
                )
            else:
                raise IOError(f"Input file is corrupted between {interval.start} and {interval.end} (in seconds)")

        if "Error initializing complex filter" in str(console_output.stdout):
            raise ValueError("Invalid render options")

        return True

    def __generate_command(self, interval_output_file: pathlib.Path, interval: Interval, apply_filter: bool):
        """
        Generates the ffmpeg command to process the video
        :param interval_output_file: Where the media interval should be saved
        :param interval: The current interval
        :param apply_filter: Whether a filter should be applied or not
        :return: ffmpeg console command
        """
        command = [
            "ffmpeg",
            "-ss", f"{interval.start}",
            "-to", f"{interval.end}",
            "-i", f"{self.__input_file}",
            "-vsync", "1",
            "-async", "1",
            "-safe", "0",
            "-ignore_unknown", "-y",
        ]

        if apply_filter:
            complex_filter = []

            if interval.is_silent:
                current_speed = self.__render_options.silent_speed
                current_volume = self.__render_options.silent_volume
            else:
                current_speed = self.__render_options.audible_speed
                current_volume = self.__render_options.audible_volume

            if not self.__render_options.audio_only:
                complex_filter.extend([
                    f"[0:v]setpts={round(1 / current_speed, 4)}*PTS[v]",
                ])

            complex_filter.extend([
                f"[0:a]atempo={round(current_speed, 4)},volume={current_volume}[a]",
            ])

            command.extend(
                ["-filter_complex", ";".join(complex_filter)]
            )

            if not self.__render_options.audio_only:
                command.extend(["-map", "[v]"])

            command.extend(["-map", "[a]"])
        else:
            if self.__render_options.audio_only:
                command.append("-vn")

        command.append(str(interval_output_file))

        return command



from unsilence.lib.intervals.Interval import Interval


class Intervals:
    """
    Collection of lib.Intervals.Interval
    """

    def __init__(self, interval_list: list = None):
        """
        Initializes a new Interval Collection
        :param interval_list: list of intervals, optional
        """
        if interval_list is None:
            interval_list = []

        self.__interval_list = interval_list

    def add_interval(self, interval):
        """
        Adds an interval to the collection
        :param interval: interval to be added
        :return: None
        """
        self.__interval_list.append(interval)

    @property
    def intervals(self):
        """
        Returns the list of intervals
        :return:
        """
        return self.__interval_list

    def optimize(self, short_interval_threshold=0.3, stretch_time=0.25):
        """
        Optimizes the Intervals to be a better fit for media cutting
        :param short_interval_threshold: The shortest allowed interval length (in seconds)
        :param stretch_time: The time that should be added/removed from a audible/silent interval
        :return: None
        """
        self.__combine_intervals(short_interval_threshold)
        self.__enlarge_audible_intervals(stretch_time)

    def __combine_intervals(self, short_interval_threshold):
        """
        Combines multiple intervals in order to remove intervals smaller than a threshold
        :param short_interval_threshold: Threshold for the shortest allowed interval
        :return: None
        """
        intervals = []
        current_interval = Interval(is_silent=None)

        for interval in self.__interval_list:
            if interval.duration <= short_interval_threshold or current_interval.is_silent == interval.is_silent:
                current_interval.end = interval.end

            else:
                if current_interval.is_silent is None:
                    current_interval.is_silent = interval.is_silent
                    current_interval.end = interval.end
                else:
                    intervals.append(current_interval)
                    current_interval = interval.copy()

        if current_interval.is_silent is None:
            current_interval.is_silent = False

        intervals.append(current_interval)

        self.__interval_list = intervals

    def __enlarge_audible_intervals(self, stretch_time):
        """
        Enlarges/Shrinks intervals based on if they are silent or audible
        :param stretch_time: Time the intervals should be enlarged/shrunken
        :return: None
        """
        for i, interval in enumerate(self.__interval_list):
            interval.enlarge_audible_interval(
                stretch_time,
                is_start_interval=(i == 0),
                is_end_interval=(i == len(self.__interval_list) - 1)
            )

    def remove_short_intervals_from_start(self, audible_speed=1, silent_speed=2):
        """
        Removes Intervals from start that are shorter than 0.5 seconds after
        speedup to avoid having a final output without an audio track
        :param audible_speed: The speed at which the audible intervals get played back at (float)
        :param silent_speed: The speed at which the silent intervals get played back at (float)
        :return: The new, possibly shorter, Intervals object
        """
        for i, interval in enumerate(self.__interval_list):
            if interval.is_silent:
                speed = silent_speed
            else:
                speed = audible_speed

            if interval.duration / speed > 0.5:
                return Intervals(self.__interval_list[i:])

        raise Exception("No interval has a length over 0.5 seconds after speed changes! This is required.")

    def copy(self):
        """
        Creates a deep copy
        :return: Deep copy of Intervals
        """
        new_interval_list = [interval.copy() for interval in self.__interval_list]

        return Intervals(new_interval_list)

    def serialize(self):
        """
        Serializes this collection
        :return: Serialized list
        """
        return [interval.serialize() for interval in self.__interval_list]

    @staticmethod
    def deserialize(serialized_obj):
        """
        Deserializes a previously serialized object and creates a new Instance from it
        :param serialized_obj: Serialized list
        :return: New instance of Intervals
        """
        interval_list = [Interval.deserialize(serialized_interval) for serialized_interval in serialized_obj]
        return Intervals(interval_list)

    def __repr__(self):
        """
        String representation
        :return: String representation
        """
        return str(self.__interval_list)



from rich.console import Console


def choice_dialog(console: Console, message: str, default: bool = None):
    """
    A simple yes/no console dialog option
    :param console: rich.Console Instance
    :param message: Message should be asked
    :param default: Default value when enter is pressed without an input (None, True, False)
    :return: Answer (True, False)
    """
    yes = "y" if default is not True else "[green]Y[/green]"
    no = "n" if default is not False else "[red]N[/red]"

    result = None
    while result is None:
        choice = console.input(f"[[[green]?[/green]]] {message} ({yes}/{no}) ").lower()
        if choice in ["yes", "y"]:
            result = True
        elif choice in ["no", "n"]:
            result = False
        elif choice == "":
            if default is not None:
                result = default

        if result is None:
            console.print("[red]Invalid input, please try again[/red]")

    return result



import queue
import shutil
import subprocess
import threading
import time
import uuid
from pathlib import Path
from types import SimpleNamespace

from unsilence.lib.intervals.Intervals import Intervals
from unsilence.lib.render_media.RenderIntervalThread import RenderIntervalThread


class MediaRenderer:
    """
    The Media Renderer handles the rendering of Intervals objects, so it processes the complete video and concatenates
    the different intervals at the end
    """

    def __init__(self, temp_path: Path):
        """
        Initializes a new MediaRenderer Object
        :param temp_path: The temp path where all temporary files should be stored
        """
        self.__temp_path = Path(temp_path).absolute()

    def render(self, input_file: Path, output_file: Path, intervals: Intervals, **kwargs):
        """
        Renders an input_file and writes the final output to output_file
        :param input_file: The file that should be processed
        :param output_file: Where the processed file should be saved
        :param intervals: The Intervals that should be processed
        :param kwargs: Keyword Args, see below
        :return: None

        kwargs:
            audio_only: Whether the output should be audio only (bool)
            audible_speed: The speed at which the audible intervals get played back at (float)
            silent_speed: The speed at which the silent intervals get played back at (float)
            audible_volume: The volume at which the audible intervals get played back at (float)
            silent_volume: The volume at which the silent intervals get played back at (float)
            drop_corrupted_intervals: Whether corrupted video intervals should be discarded or tried to recover (bool)
            threads: Number of threads to render simultaneously (int > 0)
            on_render_progress_update: Function that should be called on render progress update
                (called like: func(current, total))
            on_concat_progress_update: Function that should be called on concat progress update
                (called like: func(current, total))
        """
        input_file = Path(input_file).absolute()
        output_file = Path(output_file).absolute()

        if not input_file.exists():
            raise FileNotFoundError(f"Input file {input_file} does not exist!")

        render_options = SimpleNamespace(
            audio_only=kwargs.get("audio_only", False),
            audible_speed=kwargs.get("audible_speed", 1),
            silent_speed=kwargs.get("silent_speed", 6),
            audible_volume=kwargs.get("audible_volume", 1),
            silent_volume=kwargs.get("silent_volume", 0.5),
            drop_corrupted_intervals=kwargs.get("drop_corrupted_intervals", False)
        )

        intervals = intervals.remove_short_intervals_from_start(
            render_options.audible_speed,
            render_options.silent_speed
        )

        video_temp_path = self.__temp_path / str(uuid.uuid4())
        video_temp_path.mkdir(parents=True)

        concat_file = video_temp_path / "concat_list.txt"
        final_output = video_temp_path / f"out_final{output_file.suffix}"

        file_list = []

        thread_lock = threading.Lock()
        task_queue = queue.Queue()
        thread_list = []
        completed_tasks = []
        corrupted_intervals = []

        def handle_thread_completed_task(completed_task, corrupted):
            """
            Nested function that is called when a thread completes it current task
            :param completed_task: The completed task
            :param corrupted: If the task contained a corrupted media part
            :return: None
            """
            func = kwargs.get("on_render_progress_update", None)

            thread_lock.acquire()

            if not corrupted:
                completed_tasks.append(completed_task)
                if func is not None:
                    func(len(completed_tasks), len(intervals.intervals))
            else:
                corrupted_intervals.append(completed_task)

            thread_lock.release()

        for i in range(kwargs.get("threads", 2)):
            thread = RenderIntervalThread(i, input_file, render_options, task_queue, thread_lock,
                                          on_task_completed=handle_thread_completed_task)
            thread.start()
            thread_list.append(thread)

        for i, interval in enumerate(intervals.intervals):
            current_file_name = f"out_{i}{output_file.suffix}"
            current_path = video_temp_path / current_file_name

            file_list.append(current_path)

            task = SimpleNamespace(task_id=i, interval_output_file=current_path, interval=interval)

            thread_lock.acquire()
            task_queue.put(task)
            thread_lock.release()

        while len(completed_tasks) < (len(intervals.intervals) - len(corrupted_intervals)):
            time.sleep(0.5)

        for thread in thread_list:
            thread.stop()

        completed_file_list = [task.interval_output_file for task in sorted(completed_tasks, key=lambda x: x.task_id)]

        MediaRenderer.__concat_intervals(
            completed_file_list,
            concat_file,
            final_output,
            kwargs.get("on_concat_progress_update", None)
        )

        shutil.move(final_output, output_file)
        shutil.rmtree(video_temp_path)

    @staticmethod
    def __concat_intervals(file_list: list, concat_file: Path, output_file: Path, update_concat_progress):
        """
        Concatenates all interval files to create a finished file
        :param file_list: List of interval files
        :param concat_file: Where the ffmpeg concat filter file should be saved
        :param output_file: Where the final output file should be saved
        :param update_concat_progress: A function that is called when a step is finished
            (called like function(current, total))
        :return: None
        """
        total_files = len(file_list)

        with open(str(concat_file), "w+") as file:
            lines = [f"file {interval_file.name}\n" for interval_file in file_list]
            file.writelines(lines)

        command = [
            "ffmpeg",
            "-f", "concat",
            "-safe", "0",
            "-i", f"{concat_file.as_posix()}",
            "-c", "copy",
            "-y",
            "-loglevel", "verbose",
            f"{output_file.as_posix()}"
        ]

        console_output = subprocess.Popen(
            command,
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT,
            universal_newlines=True
        )

        current_file = 0
        for line in console_output.stdout:
            if "Auto-inserting" in line:
                if update_concat_progress is not None:
                    current_file += 1
                    update_concat_progress(current_file, total_files)
